from ._utility import read_name, write_name, read_short, write_short

class DnsRecord:
    def __new__(cls, *args, **kwargs):
        if cls is DnsRecord:
            raise Exception('Cannot instantiate abstract class {0}'.format(cls.__name__))
        else:
            return super().__new__(cls)
    def __init__(self, name, type_, class_):
        '''
        :param name: The record domain name.
        :type name: str
        :param type_: The record :class:`DnsType`.
        :type type_: int
        :param class_: The record :class:`DnsClass`.
        :type class_: int

        Represent a DNS record and allows to maipulate its properties.
        '''
        self._name = name
        self._type = type_
        self._class = class_
    #properties
    @property
    def name(self):
        '''
        The record domain name.
        '''
        return self._name
    @name.setter
    def name(self, value):
        self._name = value
    @property
    def record_type(self):
        '''
        The record :class:`DnsType`.
        '''
        return self._type
    @record_type.setter
    def record_type(self, value):
        self._type = value & 0xFFFF
    @property
    def record_class(self):
        '''
        The record :class:`DnsClass`.
        '''
        return self._class & 0x7FFF
    @record_class.setter
    def record_class(self, value):
        self._class &= 0x8000
        self._class |= (value & 0x7FFF)
    #equality operator
    def __eq__(self, other):
        if not isinstance(other, DnsRecord):
            return False
        elif (self.name == other.name and 
              self.record_type == other.record_type and 
              self.record_class == other.record_class):
            return True
        else:
            return False
    #methods
    @classmethod
    def _decode(cls, data, idx):
        name = read_name(data, idx)
        type_ = read_short(data, idx)
        class_ = read_short(data, idx)
        return name, type_, class_
    def _encode(self, data, name_compression_dictionary):
        write_name(self.name, data, name_compression_dictionary)
        write_short(self._type, data)
        write_short(self._class, data)



import unittest
from anubi.mdns import DnsPacket, DnsQRecord, DnsType, DnsClass, DnsRRecordA, DnsRRecordNotImplemented, OpCode, RCode

class EncodeTests(unittest.TestCase):  

    def test_query_1(self):
        #prepare the packet
        q_packet = DnsPacket(True)        
        q_packet.id = 0
        q_packet.AA = True
        q_packet.RD = True
        q_packet.question_records.append(DnsQRecord('test.local.', DnsType.A, DnsClass.IN))
        q_packet.question_records.append(DnsQRecord('_service.test.local.', DnsType.ANY, DnsClass.IN))
        q_packet.question_records.append(DnsQRecord('test._tcp.local.', DnsType.TXT, DnsClass.IN))
        #test encoding
        self.assertEqual(q_packet.encode(), b'\x00\x00\x05\x00\x00\x03\x00\x00\x00\x00\x00\x00\x04test\x05local\x00\x00\x01\x00\x01\x08_service\xc0\x0c\x00\xff\x00\x01\x04test\x04_tcp\xc0\x11\x00\x10\x00\x01')
    def test_query_2(self):
        #prepare the packet
        q_packet = DnsPacket(True)
        q_packet.id = 38784
        q = DnsQRecord('test.local.', DnsType.A, DnsClass.IN)
        q.prefer_unicast = True
        q_packet.question_records.append(q)
        q = DnsQRecord('_service.test.local.', DnsType.ANY, DnsClass.ANY)
        q.prefer_unicast = True
        q_packet.question_records.append(q)
        #test encoding
        self.assertEqual(q_packet.encode(False), b'\x97\x80\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x04test\x05local\x00\x00\x01\x80\x01\x08_service\x04test\x05local\x00\x00\xff\x80\xff')
    def test_answer_1(self):
        #prepare the packet
        a_packet = DnsPacket(False)
        a_packet.id = 0
        a_packet.AA = True
        a_packet.TC = True
        a_packet.answer_records.append(DnsRRecordA('test.local.', 120, '127.0.0.1'))
        r = DnsRRecordNotImplemented('test.local.', DnsType.AAAA, DnsClass.IN, 120, 16, int.to_bytes(0, 16, byteorder='big'))
        r.flush_cache = True
        a_packet.additional_records.append(r)
        #test encoding
        self.assertEqual(a_packet.encode(), b'\x00\x00\x86\x00\x00\x00\x00\x01\x00\x00\x00\x01\x04test\x05local\x00\x00\x01\x00\x01\x00\x00\x00x\x00\x04\x7f\x00\x00\x01\xc0\x0c\x00\x1c\x80\x01\x00\x00\x00x\x00\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')
    def test_answer_2(self):
        #prepare the packet
        a_packet = DnsPacket(False)
        a_packet.id = 38784
        a_packet.RA = True
        a_packet.OpCode = OpCode.NOTIFY
        a_packet.RCODE = RCode.REFUSED
        a_packet.question_records.append(DnsQRecord('test.local.', DnsType.A, DnsClass.IN))
        a_packet.question_records.append(DnsQRecord('_service.test.local.', DnsType.ANY, DnsClass.ANY))
        a_packet.answer_records.append(DnsRRecordNotImplemented('test._tcp.local.', DnsType.TXT, DnsClass.CS, 120, 8 , b'test1234'))
        a_packet.answer_records.append(DnsRRecordA('test.local.', 60, '127.0.0.2'))
        a_packet.answer_records.append(DnsRRecordA('test.local.', 240, '127.0.0.3'))
        a_packet.authority_records.append(DnsRRecordNotImplemented('test.local.', DnsType.NS, DnsClass.IN, 1000, 18 , b'test1.test2.local.'))
        a_packet.additional_records.append(DnsRRecordNotImplemented('test.local.', DnsType.AAAA, DnsClass.IN, 120, 16, int.to_bytes(0, 16, byteorder='big')))
        a_packet.additional_records.append(DnsRRecordNotImplemented('_service.test.local.', DnsType.SRV, DnsClass.IN, 4500, 18, b'\x00\x00\x00\x00\x00\x15\x09_service1\xc0\x0c'))
        #test encoding
        self.assertEqual(a_packet.encode(), b'\x97\x80\xa0\x85\x00\x02\x00\x03\x00\x01\x00\x02\x04test\x05local\x00\x00\x01\x00\x01\x08_service\xc0\x0c\x00\xff\x00\xff\x04test\x04_tcp\xc0\x11\x00\x10\x00\x02\x00\x00\x00x\x00\x08test1234\xc0\x0c\x00\x01\x00\x01\x00\x00\x00<\x00\x04\x7f\x00\x00\x02\xc0\x0c\x00\x01\x00\x01\x00\x00\x00\xf0\x00\x04\x7f\x00\x00\x03\xc0\x0c\x00\x02\x00\x01\x00\x00\x03\xe8\x00\x12test1.test2.local.\xc0\x0c\x00\x1c\x00\x01\x00\x00\x00x\x00\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xc0\x1c\x00!\x00\x01\x00\x00\x11\x94\x00\x12\x00\x00\x00\x00\x00\x15\t_service1\xc0\x0c')


import unittest
from anubi.mdns import DnsPacket, DnsType, DnsClass, OpCode, RCode

class DecodeTests(unittest.TestCase):
    
    @classmethod
    def setUpClass(cls):
        #decode known packets
        cls._q_packet_1 = DnsPacket.decode(b'\x00\x00\x05\x00\x00\x03\x00\x00\x00\x00\x00\x00\x04test\x05local\x00\x00\x01\x00\x01\x08_service\xc0\x0c\x00\xff\x00\x01\x04test\x04_tcp\xc0\x11\x00\x10\x00\x01')
        cls._q_packet_2 = DnsPacket.decode(b'\x97\x80\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x04test\x05local\x00\x00\x01\x80\x01\x08_service\x04test\x05local\x00\x00\xff\x80\xff')
        cls._a_packet_1 = DnsPacket.decode(b'\x00\x00\x86\x00\x00\x00\x00\x01\x00\x00\x00\x01\x04test\x05local\x00\x00\x01\x00\x01\x00\x00\x00x\x00\x04\x7f\x00\x00\x01\xc0\x0c\x00\x1c\x80\x01\x00\x00\x00x\x00\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')
        cls._a_packet_2 = DnsPacket.decode(b'\x97\x80\xa0\x85\x00\x02\x00\x03\x00\x01\x00\x02\x04test\x05local\x00\x00\x01\x00\x01\x08_service\xc0\x0c\x00\xff\x00\xff\x04test\x04_tcp\xc0\x11\x00\x10\x00\x02\x00\x00\x00x\x00\x08test1234\xc0\x0c\x00\x01\x00\x01\x00\x00\x00<\x00\x04\x7f\x00\x00\x02\xc0\x0c\x00\x01\x00\x01\x00\x00\x00\xf0\x00\x04\x7f\x00\x00\x03\xc0\x0c\x00\x02\x00\x01\x00\x00\x03\xe8\x00\x12test1.test2.local.\xc0\x0c\x00\x1c\x00\x01\x00\x00\x00x\x00\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xc0\x1c\x00!\x00\x01\x00\x00\x11\x94\x00\x12\x00\x00\x00\x00\x00\x15\t_service1\xc0\x0c')

    def test_id(self):
        self.assertEqual(DecodeTests._q_packet_1.id, 0)
        self.assertEqual(DecodeTests._q_packet_2.id, 38784)
        self.assertEqual(DecodeTests._a_packet_1.id, 0)
        self.assertEqual(DecodeTests._a_packet_2.id, 38784)
    def test_query_bit(self):
        self.assertTrue(DecodeTests._q_packet_1.QR)
        self.assertTrue(DecodeTests._q_packet_2.QR)
        self.assertFalse(DecodeTests._a_packet_1.QR)
        self.assertFalse(DecodeTests._a_packet_2.QR)
    def test_op_code(self):
        self.assertEqual(DecodeTests._q_packet_1.OpCode, 0)
        self.assertEqual(DecodeTests._q_packet_2.OpCode, 0)
        self.assertEqual(DecodeTests._a_packet_1.OpCode, 0)
        self.assertEqual(DecodeTests._a_packet_2.OpCode, OpCode.NOTIFY)
    def test_authoritative_answer_bit(self):
        self.assertTrue(DecodeTests._q_packet_1.AA)
        self.assertFalse(DecodeTests._q_packet_2.AA)
        self.assertTrue(DecodeTests._a_packet_1.AA)
        self.assertFalse(DecodeTests._a_packet_2.AA)
    def test_truncated_bit(self):
        self.assertFalse(DecodeTests._q_packet_1.TC)
        self.assertFalse(DecodeTests._q_packet_2.TC)
        self.assertTrue(DecodeTests._a_packet_1.TC)
        self.assertFalse(DecodeTests._a_packet_2.TC)
    def test_recursion_desired_bit(self):
        self.assertTrue(DecodeTests._q_packet_1.RD)
        self.assertFalse(DecodeTests._q_packet_2.RD)
        self.assertFalse(DecodeTests._a_packet_1.RD)
        self.assertFalse(DecodeTests._a_packet_2.RD)
    def test_recursion_avaialble_bit(self):
        self.assertFalse(DecodeTests._q_packet_1.RA)
        self.assertFalse(DecodeTests._q_packet_2.RA)
        self.assertFalse(DecodeTests._a_packet_1.RA)
        self.assertTrue(DecodeTests._a_packet_2.RA)
    def test_authenticated_data_bit(self):
        self.assertFalse(DecodeTests._q_packet_1.AD)
        self.assertFalse(DecodeTests._q_packet_2.AD)
        self.assertFalse(DecodeTests._a_packet_1.AD)
        self.assertFalse(DecodeTests._a_packet_2.AD)
    def test_checking_disabled_bit(self):
        self.assertFalse(DecodeTests._q_packet_1.CD)
        self.assertFalse(DecodeTests._q_packet_1.CD)
        self.assertFalse(DecodeTests._a_packet_1.CD)
        self.assertFalse(DecodeTests._a_packet_2.CD)
    def test_rcode(self):
        self.assertEqual(DecodeTests._q_packet_1.RCODE, 0)
        self.assertEqual(DecodeTests._q_packet_2.RCODE, 0)
        self.assertEqual(DecodeTests._a_packet_1.RCODE, 0)
        self.assertEqual(DecodeTests._a_packet_2.RCODE, RCode.REFUSED)
    def test_question_records_count(self):
        self.assertEqual(len(DecodeTests._q_packet_1.question_records), 3)
        self.assertEqual(len(DecodeTests._q_packet_2.question_records), 2)
        self.assertEqual(len(DecodeTests._a_packet_1.question_records), 0)
        self.assertEqual(len(DecodeTests._a_packet_2.question_records), 2)
    def test_answer_records_count(self):
        self.assertEqual(len(DecodeTests._q_packet_1.answer_records), 0)
        self.assertEqual(len(DecodeTests._q_packet_2.answer_records), 0)
        self.assertEqual(len(DecodeTests._a_packet_1.answer_records), 1)
        self.assertEqual(len(DecodeTests._a_packet_2.answer_records), 3)
    def test_authority_records_count(self):
        self.assertEqual(len(DecodeTests._q_packet_1.authority_records), 0)
        self.assertEqual(len(DecodeTests._q_packet_2.authority_records), 0)
        self.assertEqual(len(DecodeTests._a_packet_1.authority_records), 0)
        self.assertEqual(len(DecodeTests._a_packet_2.authority_records), 1)
    def test_additional_records_count(self):
        self.assertEqual(len(DecodeTests._q_packet_1.additional_records), 0)
        self.assertEqual(len(DecodeTests._q_packet_2.additional_records), 0)
        self.assertEqual(len(DecodeTests._a_packet_1.additional_records), 1)
        self.assertEqual(len(DecodeTests._a_packet_2.additional_records), 2)
    def test_resource_record_A(self):
        self.assertEqual(sum(1 for r in DecodeTests._a_packet_1.answer_records if r.record_type == DnsType.A), 1)
        self.assertEqual(DecodeTests._a_packet_1.answer_records[0].record_type, DnsType.A)
        self.assertEqual(DecodeTests._a_packet_1.answer_records[0].record_class, DnsClass.IN)
        self.assertEqual(DecodeTests._a_packet_1.answer_records[0].ttl, 120)
        self.assertEqual(DecodeTests._a_packet_1.answer_records[0].address, '127.0.0.1')
        self.assertEqual(sum(1 for r in DecodeTests._a_packet_2.answer_records if r.record_type == DnsType.A), 2)
        self.assertEqual(DecodeTests._a_packet_2.answer_records[1].record_type, DnsType.A)
        self.assertEqual(DecodeTests._a_packet_2.answer_records[1].record_class, DnsClass.IN)
        self.assertEqual(DecodeTests._a_packet_2.answer_records[1].ttl, 60)
        self.assertEqual(DecodeTests._a_packet_2.answer_records[1].address, '127.0.0.2')
        self.assertEqual(DecodeTests._a_packet_2.answer_records[2].record_type, DnsType.A)
        self.assertEqual(DecodeTests._a_packet_2.answer_records[2].record_class, DnsClass.IN)
        self.assertEqual(DecodeTests._a_packet_2.answer_records[2].ttl, 240)
        self.assertEqual(DecodeTests._a_packet_2.answer_records[2].address, '127.0.0.3')
    def test_record_names(self):
        self.assertEqual(DecodeTests._q_packet_1.question_records[0].name, 'test.local.')
        self.assertEqual(DecodeTests._q_packet_1.question_records[1].name, '_service.test.local.')
        self.assertEqual(DecodeTests._q_packet_1.question_records[2].name, 'test._tcp.local.')
        self.assertEqual(DecodeTests._q_packet_2.question_records[0].name, 'test.local.')
        self.assertEqual(DecodeTests._q_packet_2.question_records[1].name, '_service.test.local.')
        self.assertEqual(DecodeTests._a_packet_1.answer_records[0].name, 'test.local.')
        self.assertEqual(DecodeTests._a_packet_1.additional_records[0].name, 'test.local.')
        self.assertEqual(DecodeTests._a_packet_2.question_records[0].name, 'test.local.')
        self.assertEqual(DecodeTests._a_packet_2.question_records[1].name, '_service.test.local.')
        self.assertEqual(DecodeTests._a_packet_2.answer_records[0].name, 'test._tcp.local.')
        self.assertEqual(DecodeTests._a_packet_2.answer_records[1].name, 'test.local.')
        self.assertEqual(DecodeTests._a_packet_2.answer_records[2].name, 'test.local.')
        self.assertEqual(DecodeTests._a_packet_2.authority_records[0].name, 'test.local.')
        self.assertEqual(DecodeTests._a_packet_2.additional_records[0].name, 'test.local.')
        self.assertEqual(DecodeTests._a_packet_2.additional_records[1].name, '_service.test.local.')
    def test_record_types(self):
        self.assertEqual(DecodeTests._q_packet_1.question_records[0].record_type, DnsType.A)
        self.assertEqual(DecodeTests._q_packet_1.question_records[1].record_type, DnsType.ANY)
        self.assertEqual(DecodeTests._q_packet_1.question_records[2].record_type, DnsType.TXT)
        self.assertEqual(DecodeTests._q_packet_2.question_records[0].record_type, DnsType.A)
        self.assertEqual(DecodeTests._q_packet_2.question_records[1].record_type, DnsType.ANY)
        self.assertEqual(DecodeTests._a_packet_1.answer_records[0].record_type, DnsType.A)
        self.assertEqual(DecodeTests._a_packet_1.additional_records[0].record_type, DnsType.AAAA)
        self.assertEqual(DecodeTests._a_packet_2.question_records[0].record_type, DnsType.A)
        self.assertEqual(DecodeTests._a_packet_2.question_records[1].record_type, DnsType.ANY)
        self.assertEqual(DecodeTests._a_packet_2.answer_records[0].record_type, DnsType.TXT)
        self.assertEqual(DecodeTests._a_packet_2.answer_records[1].record_type, DnsType.A)
        self.assertEqual(DecodeTests._a_packet_2.answer_records[2].record_type, DnsType.A)
        self.assertEqual(DecodeTests._a_packet_2.authority_records[0].record_type, DnsType.NS)
        self.assertEqual(DecodeTests._a_packet_2.additional_records[0].record_type, DnsType.AAAA)
        self.assertEqual(DecodeTests._a_packet_2.additional_records[1].record_type, DnsType.SRV)
    def test_record_classes(self):
        self.assertEqual(DecodeTests._q_packet_1.question_records[0].record_class, DnsClass.IN)
        self.assertEqual(DecodeTests._q_packet_1.question_records[1].record_class, DnsClass.IN)
        self.assertEqual(DecodeTests._q_packet_1.question_records[2].record_class, DnsClass.IN)
        self.assertEqual(DecodeTests._q_packet_2.question_records[0].record_class, DnsClass.IN)
        self.assertEqual(DecodeTests._q_packet_2.question_records[1].record_class, DnsClass.ANY)
        self.assertEqual(DecodeTests._a_packet_1.answer_records[0].record_class, DnsClass.IN)
        self.assertEqual(DecodeTests._a_packet_1.additional_records[0].record_class, DnsClass.IN)
        self.assertEqual(DecodeTests._a_packet_2.question_records[0].record_class, DnsClass.IN)
        self.assertEqual(DecodeTests._a_packet_2.question_records[1].record_class, DnsClass.ANY)
        self.assertEqual(DecodeTests._a_packet_2.answer_records[0].record_class, DnsClass.CS)
        self.assertEqual(DecodeTests._a_packet_2.answer_records[1].record_class, DnsClass.IN)
        self.assertEqual(DecodeTests._a_packet_2.answer_records[2].record_class, DnsClass.IN)
        self.assertEqual(DecodeTests._a_packet_2.authority_records[0].record_class, DnsClass.IN)
        self.assertEqual(DecodeTests._a_packet_2.additional_records[0].record_class, DnsClass.IN)
        self.assertEqual(DecodeTests._a_packet_2.additional_records[1].record_class, DnsClass.IN)
    def test_prefer_unicast_bit(self):
        self.assertFalse(any(map(lambda r: r.prefer_unicast, DecodeTests._q_packet_1.question_records)))
        self.assertTrue(all(map(lambda r: r.prefer_unicast, DecodeTests._q_packet_2.question_records)))
        self.assertFalse(any(map(lambda r: r.prefer_unicast, DecodeTests._a_packet_2.question_records)))
    def test_flush_cache_bit(self):
        self.assertFalse(DecodeTests._a_packet_1.answer_records[0].flush_cache)
        self.assertTrue(DecodeTests._a_packet_1.additional_records[0].flush_cache)
        self.assertFalse(any(map(lambda r: r.flush_cache ,DecodeTests._a_packet_2.answer_records)))
        self.assertFalse(any(map(lambda r: r.flush_cache ,DecodeTests._a_packet_2.authority_records)))
        self.assertFalse(any(map(lambda r: r.flush_cache ,DecodeTests._a_packet_2.additional_records)))


class Index():
    def __init__(self, value = 0):
        '''
        :param value: The initial value of the instance. Default to 0.
        :type value: int or :class:`Index`

        Represent a mutable integer that is very usefull when used as index in encode/decode methods since it will be automatically updated and so keeps track of the current position in the data array.
        this class mimic the `int` builtin class for addition and indexing.
        '''
        if isinstance(value, Index):
            self._value = value._value
        else:
            self._value = value
    def __index__(self):
        return self._value
    def __add__(self, other):
        return Index(self._value + other)
    def __radd__(self, other):
        return Index(self._value + other)
    def advance(self, quantity):
        '''
        :param quantity: The ammount to which the instance value has to be incremented.
        :type quantity: int

        Increment the instance value by the given quantity.
        '''
        self._value += quantity

def read_short(data, idx):
    res =  int.from_bytes(data[idx: idx + 2], byteorder='big')
    idx.advance(2)
    return res
def read_int(data, idx):
    res =  int.from_bytes(data[idx: idx + 4], byteorder='big')
    idx.advance(4)
    return res
def write_short(num, data):
    data.extend(num.to_bytes(2, byteorder='big'))
def write_int(num, data):
    data.extend(num.to_bytes(4, byteorder='big'))
def read_name(data, idx):
    size = data[idx]
    if size == 0:
        name = ''
        idx.advance(1)
    elif (size & 0xC0) != 0:
        #name is compressed
        tmp_idx = Index()
        tmp_idx = Index(read_short([size & 0x3F, data[idx + 1]], tmp_idx))
        name = read_name(data, tmp_idx)
        idx.advance(2)
    else:
        #name is not compressed
        idx.advance(1)
        name = data[idx:idx + size].decode('UTF-8') + '.'
        idx.advance(size)
        n = read_name(data, idx)
        name += n
    return name
def write_name(name, data, name_compression_dictionary):
    tokens = list(map(lambda t: bytes([len(t)]) + t ,[t.encode('UTF-8') for t in name.split('.') if len(t) != 0]))
    for token in [(tokens[i], b''.join(tokens[i:])) for i in range(len(tokens))]:
        if name_compression_dictionary != None and token[1] in name_compression_dictionary:
            offset = name_compression_dictionary[token[1]]
            offset |= 0xC000
            write_short(offset, data)
            return
        else:
            if name_compression_dictionary != None:
                name_compression_dictionary[token[1]] = len(data)
            data.extend(token[0])
    data.append(0)


import socket
import threading
import queue
import random
import errno
import logging
import time
import select
import os
from .dnspacket import DnsPacket
from .dnsenums import OpCode, RCode

logger = logging.getLogger(__name__)

class QueueRecord():
    def __init__(self, query, answers, unicast_address):
        self.query = query
        self.answers = answers
        self.unicast_address = unicast_address

class DnsReplySender(threading.Thread):
    def __init__(self, sock, multicast_address):
        super().__init__(name='DnsReplySender')
        self.__reply_queue = queue.Queue(-1) #infinite queue
        self.__reply_queue_lock = threading.Semaphore(0)
        self.__sock = sock
        self.__multicast_address = multicast_address

    def run(self):
        try:
            while True:
                #get item from queue
                self.__reply_queue_lock.acquire()
                data = self.__reply_queue.get_nowait()
                if data is None:
                    logger.info('Exiting message sender thread')
                    #stop the thread
                    return
                received_from_multicast = data.unicast_address[1] == self.__multicast_address[1]
                prefer_unicast = any(map(lambda q: q.prefer_unicast, data.query.question_records))
                if prefer_unicast or (not received_from_multicast):
                    #send unicast message
                    unicast_reply = DnsPacket(False)
                    unicast_reply.id = data.query.id
                    unicast_reply.question_records.extend(data.query.question_records)
                    unicast_reply.answer_records.extend(data.answers)
                    logger.info('Unicast: sending %s to %s', unicast_reply, data.unicast_address)
                    self.__sock.sendto(unicast_reply.encode(), data.unicast_address)
                if (not prefer_unicast) or received_from_multicast:
                    #send multicast message
                    multicast_reply = DnsPacket(False)
                    multicast_reply.answer_records.extend(data.answers)
                    delay = random.randint(20, 120) / 1000.0
                    #multicast reply must be delayed by a random ammount of ms to avoid collisions
                    logger.info('Multicast: sending %s', multicast_reply)
                    time.sleep(delay)
                    self.__sock.sendto(multicast_reply.encode(), self.__multicast_address)
        except Exception as ex:
            logger.error('Exiting message sender thread with error {0}'.format(ex))
            return

    def send(self, query, answers, unicast_address):
        record = QueueRecord(query, answers, unicast_address)
        self.__reply_queue.put(record)
        self.__reply_queue_lock.release()

    def stop(self):
        #token to stop the thread
        self.__reply_queue.put(None)
        #notify that new data are available
        self.__reply_queue_lock.release()
        logger.info('Stopping message sender thread')


class mDNS(threading.Thread):
    __MDNS_ADDRESS = ('224.0.0.251', 5353)
    __MAX_PACKET_SIZE = 9000 - 20 - 8#Max packet - IPv4 header - UDP header

    def __init__(self):
        '''
        Implements the multicast DNS logic.
        Once started it spawn a new thread that reply to mDNS queries when there is at least one answer
        to the query question in the internal record database.
        '''
        super().__init__(name='mDNS responder')
        #record database
        self.__records_db = []
        #open the socket
        self.__sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        #set socket options
        self.__sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if hasattr(socket, 'SO_REUSEPORT'):
            self.__sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        self.__sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 1)
        self.__sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_LOOP, True)
        #register to multicast group
        mGroup = socket.inet_aton(self.__MDNS_ADDRESS[0]) + socket.INADDR_ANY.to_bytes(4, byteorder='big')
        self.__sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mGroup)
        #bind the socket to the correct port
        self.__sock.bind(('', self.__MDNS_ADDRESS[1]))
        #create pipes used for stopping thread
        self.__read_pipe, self.__write_pipe = os.pipe()
        #reply sender
        self.__reply_sender = DnsReplySender(self.__sock, self.__MDNS_ADDRESS)

    def run(self):
        self.__reply_sender.start()
        while True:
            try:
                ready_read, _, _ = select.select([self.__sock, self.__read_pipe], [], [])
                if (self.__sock in ready_read):
                    data, addr = self.__sock.recvfrom(4096)
                    logger.debug('Received request %s', data.hex())
                    try:
                        query = DnsPacket.decode(data)
                    except:
                        logger.warning('Fail to decode received data')
                        continue
                    #check if we have some record that ask questions
                    answers = []
                    #process only query with OP CODE: QUERY and R CODE: NOERROR
                    if query.OpCode == OpCode.QUERY and query.RCODE == RCode.NOERROR:
                        for question in query.question_records:
                            answers += list(filter(lambda r: r.answer_to(question), self.__records_db))
                        if len(answers) != 0:
                            if self.__reply_sender.is_alive():
                                #we have some answers, prepare the packet to send back
                                self.__reply_sender.send(query, answers, addr)
                            else:
                                logger.error('Exiting mDNS thread because reply sender thread stop working')
                                return
                else:
                    logger.info('Exiting mDNS thread')
                    self.__sock.close()
                    os.close(self.__read_pipe)
                    return
            except Exception as ex:
                logger.error('Exiting mDNS thread with error {0}'.format(ex))
                #error occurred, exit
                self.__sock.close()
                return
    def start(self):
        '''
        Start the mDSN instance. Once started the instance reply to mDNS queries if the internal database
        contains at least one record that reply to query questions.
        This method cannot be called on a stopped instance.
        '''
        super().start()
    def stop(self):
        '''
        Stop a running instance of mDNS that do not reply to queries anymore.
        Once stopped an mDNS instance cannot be started again.
        '''
        logger.info('Stopping mDNS thread')
        self.__reply_sender.stop()
        self.__reply_sender.join()
        w = os.fdopen(self.__write_pipe, 'w')
        w.write('\0')
        w.close()
    def add_record(self, record):
        '''
        :param record: The record that has to be added to the internal database.
        :type record: :class:`DnsRRecord`       
        :return: True if the record has been added, False otherwise.

        Add the given record to the internal class database.
        This method can be called any time, even after the instance has been started by the 
        :func:`start` method.
        '''
        #check if the record already exists
        if record in self.__records_db:
            logging.warning('Record {0} already present in record database'.format(record))
            return False
        else:
            self.__records_db.append(record)
            logger.info('Add {0} record to record database'.format(record))
            return True


import unittest
from anubi.mdns import mDNS, DnsRRecordA, DnsPacket, DnsQRecord, DnsType, DnsClass
import socket

class CommunicationTests(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls._mdns = mDNS()
        #replace the class socket with a new one connected to localhost
        cls.__mdns_address = ('127.0.0.1', mDNS._mDNS__MDNS_ADDRESS[1])
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        #set socket options
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if hasattr(socket, 'SO_REUSEPORT'):
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        sock.bind(cls.__mdns_address)
        cls._mdns._mDNS__sock.close()
        cls._mdns._mDNS__sock = sock
        cls._mdns._mDNS__reply_sender._DnsReplySender__sock = sock
        cls._mdns.add_record(DnsRRecordA('test.local.', 120, '127.0.0.1'))
        cls._mdns.start()
    @classmethod
    def tearDownClass(cls):
        cls._mdns.stop()
        cls._mdns.join()

    def setUp(self):
        self.__sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    def tearDown(self):
        self.__sock.close()

    def test_address_query_1(self):
        query = DnsPacket(True)
        question = DnsQRecord('test.local.', DnsType.A, DnsClass.IN)
        #we cannot use multicast on localhost, so force unicast reply
        question.prefer_unicast = True
        query.question_records.append(question)
        self.__sock.sendto(query.encode(), self.__mdns_address)
        data = self.__sock.recv(8192)
        reply = DnsPacket.decode(data)
        self.assertEqual(query.question_records[0], reply.question_records[0])
        self.assertEqual(reply.answer_records[0].name, 'test.local.')
        self.assertEqual(reply.answer_records[0].record_type, DnsType.A)
        self.assertEqual(reply.answer_records[0].record_class, DnsClass.IN)
        self.assertEqual(reply.answer_records[0].ttl, 120)
        self.assertEqual(reply.answer_records[0].address, '127.0.0.1')
    def test_address_query_2(self):
        query = DnsPacket(True)
        question = DnsQRecord('test.local.', DnsType.ANY, DnsClass.ANY)
        #we cannot use multicast on localhost, so force unicast reply
        question.prefer_unicast = True
        query.question_records.append(question)
        self.__sock.sendto(query.encode(), self.__mdns_address)
        data = self.__sock.recv(8192)
        reply = DnsPacket.decode(data)
        self.assertEqual(query.question_records[0], reply.question_records[0])
        self.assertEqual(reply.answer_records[0].name, 'test.local.')
        self.assertEqual(reply.answer_records[0].record_type, DnsType.A)
        self.assertEqual(reply.answer_records[0].record_class, DnsClass.IN)
        self.assertEqual(reply.answer_records[0].ttl, 120)
        self.assertEqual(reply.answer_records[0].address, '127.0.0.1')


from ._dnsrecord import DnsRecord

class DnsQRecord(DnsRecord):
    #properties
    @property
    def prefer_unicast(self):
        '''
        Flag that indicates if unicast reply are preferred (True) or not (False).
        '''
        return (self._class & 0x8000) != 0
    @prefer_unicast.setter
    def prefer_unicast(self, value):
        if value:
            self._class |= 0x8000
        else:
            self._class &= 0x7FFF
    #methods
    @classmethod
    def decode(cls, data, idx):
        '''
        :param data: Array of bytes that has to be decoded.
        :type data: bytearray
        :param idx: The index from where the record data begins in the `data` array.
        :type idx: :class:`Index`

        Decode the given data starting from the given index into a :class:`DnsQRecord` instance.
        '''
        return DnsQRecord(*super()._decode(data, idx))
    def encode(self, data, name_compression_dictionary):
        '''
        :param data: The `bytearray` to which the instance data will be appended.
        :type data: bytearray
        :param name_compression_dictionary: a dictionary used to perform name compression. Keys are the domain names, values are the indexes in the `data` array where the domain name begin.
        :type name_compression_dictionary: dict[str, int]

        Encode the intance to a `bytearray` compressing domain names according with the given `name_compression_dictionary`.
        The resulting `bytearray` will be attached to the given `data` parameter.
        If `name_compression_dictionary` is `None` no name compression will be applied.
        '''
        super()._encode(data, name_compression_dictionary)


'''
This module provides a simple DNS server that respond to muticast queries.

To use it simply 'import mdns' and create a mDNS instance,
add records to it and start it.

The server respond only on IPv4 address and support following record types:
- A: IPv4 address
'''

__version__ = '0.0.2a'
__author__ = 'Andrea Parisotto'

from .mdns import mDNS
from .dnsenums import DnsClass, DnsType, OpCode, RCode
from .dnsqrecord import DnsQRecord
from .dnsrrecords import DnsRRecord, DnsRRecordA, DnsRRecordNotImplemented, register_subclass
from .dnspacket import DnsPacket
from ._utility import Index
import logging

logging.getLogger(__name__).addHandler(logging.NullHandler())


from enum import IntEnum

class OpCode(IntEnum):
    '''
    DNS packet OpCodes.
    '''
    #: Query
    QUERY = 0
    #: Inverse Query (Obsolete)
    IQUERY = 1
    #: Status
    STATUS = 2
    #: Notify
    NOTIFY = 4
    #: Update
    UPDATE = 5
    #: DNS Stateful Operations
    DSO = 6

class RCode(IntEnum):
    '''
    DNS packet RCODEs.
    '''
    #: No error
    NOERROR = 0
    #: Format error
    FORMERR = 1
    #: Server failure
    SERVFAIL = 2
    #: Non-Existent domain
    NXDOMAIN = 3
    #: Not implemented
    NOTIMP = 4
    #: Query refused
    REFUSED = 5
    #: Name exist when it should not
    YXDOMAIN = 6
    #: RR Set exist when it should not
    YXRRSET = 7
    #: RR Set that should exist does not
    NXRRSET = 8
    #: Server not authoritative for zone/Not authorized
    NOTAUTH = 9
    #: Name not contained in zone
    NOTZONE = 10
    #: DSO-TYPE not implemented
    DSOTYPENI = 11
    #: Bad OPT version
    BADVERS = 16
    #: TSIG signature failure
    BADSIG = 16
    #: Key not recognized
    BADKEY = 17
    #: Signature out of time window
    BADTIME = 18
    #: Bad TKEY mode
    BADMODE = 19
    #: Duplicate key name
    BADNAME = 20
    #: Algorithm not supported
    BADALG = 21
    #: Bad truncation
    BADTRUNC = 22
    #: Bad/Missing server cookie
    BADCOOKIE = 23

class DnsType(IntEnum):
    '''
    DNS record types.
    '''
    #: Host address (IPv4)
    A = 1
    #: Authoritative name server
    NS = 2
    #: Mail destination (Obsolete, use :attr:`MX`)
    MD = 3
    #: Mail forwarder (Obsolete, use :attr:`MX`)
    MF = 4
    #: Canonical name for alias
    CNAME = 5
    #: Mark starts of a zone of authority
    SOA = 6
    #: Mailbox domain name (Experimental)
    MB = 7
    #: Mail group member (Experimental)
    MG = 8
    #: Mail rename domain name (Experimental)
    MR = 9
    #: Null (Experimental)
    NULL = 10
    #: Well known service description
    WKS = 11
    #: Domain name pointer
    PTR = 12
    #: Host information
    HINFO = 13
    #: Mailbox or mail list information
    MINFO = 14
    #: Mail exchange
    MX = 15
    #: Text string
    TXT = 16
    #: Responsible person
    RP = 17
    #: AFS database location
    AFSDB = 18
    #: X.25 PSDN address
    X25 = 19
    #: ISDN address
    ISDN = 20
    #: Route through
    RT = 21
    #: NSAP address (NSAP stype :attr:`A` record)
    NSAP = 22
    #: Domain name pointer, NSAP style
    NSAP_PTR = 23
    #: Security signature
    SIG = 24
    #: Security key
    KEY = 25
    #: X.400 mail mapping information
    PX = 26
    #: Geographical position
    GPOS = 27
    #: Host address (IPv6)
    AAAA = 28
    #: Location information
    LOC = 29
    #: Next domain (Obsolete)
    NXT = 30
    #: Endpoint identifier
    EID = 31
    #: Nimrod locator
    NIMLOC = 32
    #: Server selection
    SRV = 33
    #: ATM address
    ATMA = 34
    #: Naming authority pointer
    NAPTR = 35
    #: Key exchanger
    KX = 36
    #: CERT
    CERT = 37
    #: Host address (IPv6 - Obsolete, use :attr:`AAAA`)
    A6 = 38
    #: DNAME
    DNAME = 39
    #: SINK
    SINK = 40
    #: OPT
    OPT = 41
    #: APL
    APL = 42
    #: Delegation singer
    DS = 43
    #: SSH key fingerprint
    SSHFP = 44
    #: IPSECKEY
    IPSECKEY = 45
    #: RRSIG
    RRSIG = 46
    #: NSEC
    NSEC = 47
    #: DNSKEY
    DNSKEY = 48
    #: DHCID
    DHCID = 49
    #: NSEC3
    NSEC3 = 50
    #: NSEC3PARAM
    NSEC3PARAM = 51
    #: TLSA
    TLSA = 52
    #: S/MIME cer association
    SIMMEA = 53
    #: Host identity protocol
    HIP = 55
    #: NINFO
    NINFO = 56
    #: RKEY
    RKEY = 57
    #: Trust anchor link
    TALINK = 58
    #: Child DS
    CDS = 59
    #: DNSKEY(s) the child wants reflected in DS
    CDNSKEY = 60
    #: OpenPGP key
    OPENPGPKEY = 61
    #: Child-To-Parent syncronization
    CSYNC = 62
    #: Message digest for DNS zone
    ZONEMD = 63
    #: SPF
    SPF = 99
    #: UINFO (IANA reservedd)
    UINFO = 100
    #: UID (IANA reservedd)
    UID = 101
    #: GID (IANA reservedd)
    GID = 102
    #: UNSPEC (IANA reservedd)
    UNSPEC = 103
    #: NID
    NID = 104
    #: L32
    L32 = 105
    #: L64
    L64 = 106
    #: LP
    LP = 107
    #: EUI-48 address
    EUI48 = 108
    #: EUI-64 address
    EUI64 = 109
    #: Transaction key
    TKEY = 249
    #: Transaction signature
    TSIG = 250
    #: Incremental transfer
    IXFR = 251
    #: Transfer of an entire zone
    AXFR = 252
    #: Mailbox related resource record (:attr:`MB`, :attr:`MG` or :attr:`MR`)
    MAILB = 253
    #: Mail agent (Obsolete, see :attr:`MX`)
    MAILA = 254
    #: Any
    ANY = 255
    #: URI
    URI = 256
    #: Certification authority restriction
    CAA = 257
    #: Application visibility and control
    AVC = 258
    #: Digital object architecture
    DOA = 259
    #: Automatic multicast tunnelling relay
    AMTRELAY = 260
    #: DNSSEC trust authorities
    TA = 32769
    #: DNSSEC lookaside validation (Obsolete)
    DLV = 32769
    
    @classmethod
    def contains_value(cls, value):
        '''
        :param value: The value to be tested.
        :type value: int

        Test if the given value correspond to any of the enum values.
        '''
        return value in cls.__members__.values()

class DnsClass(IntEnum):
    '''
    DNS class types.
    '''
    #: Internet
    IN = 1
    #: CSNET (Obsolete)
    CS = 2
    #: Chaos
    CH = 3
    #: Hesoid
    HS = 4
    #: None
    NONE = 254
    #: Any
    ANY = 255
    
    @classmethod
    def contains_value(cls, value):
        '''
        :param value: The value to be tested.
        :type value: int

        Test if the given value correspond to any of the enum values.
        '''
        return value in cls.__members__.values()



from ._utility import Index, read_short, write_short
from .dnsqrecord import DnsQRecord
from .dnsrrecords import DnsRRecord, DnsRRecordNotImplemented, DnsRRecordA
from .dnsenums import DnsType, DnsClass
import logging

logger = logging.getLogger(__name__)

class DnsPacket:
    __QR_MASK = 0x8000
    __OPCODE_MASK = 0x7800
    __AA_MASK = 0x0400
    __TC_MASK = 0x0200
    __RD_MASK = 0x0100
    __RA_MASK = 0x0080
    __Z_MASK = 0x0040
    __AD_MASK = 0x0020
    __CD_MASK = 0x0010
    __RCODE_MASK = 0x000F
    def __init__(self, is_query):
        '''
        :param is_query: A flag that specify if the packes is a query packet or not.
        :type is_query: bool

        Represent a DNS packet and exposes API to control packet content.
        '''
        self._id = 0
        self.__flags = 0
        self.QR = is_query
        self._question_records = []
        self._answer_records = []
        self._authority_records = []
        self._additional_records = []
    #methods
    @classmethod
    def decode(cls, data):
        '''
        :param data: Array of bytes that has to be decoded.
        :type data: bytearray
        :return: A new instance of :class:`DnsPacket` built from given data.

        Decodes the given data into a :class:`DnsPacket` instance.
        '''
        instance = cls(False)
        idx = Index()
        instance._id = read_short(data, idx)
        instance.__flags = read_short(data, idx)
        #zero flag is ignored and must be alwais be zero
        instance.__flags &= ~cls.__Z_MASK
        question_count = read_short(data, idx)
        answer_count = read_short(data, idx)
        authority_count = read_short(data, idx)
        additional_conunt = read_short(data, idx)
        for i in range(question_count):
            try:
                instance.question_records.append(DnsQRecord.decode(data, idx))
            except Exception as ex:
                logger.warning('Fail to decode question record %d with error %s', i, str(ex))
        for i in range(answer_count):
            try:
                instance.answer_records.append(DnsRRecord.decode(data, idx))
            except Exception as ex:
                logger.warning('Fail to decode answer record %d with error %s', i, str(ex))
        for i in range(authority_count):
            try:
                instance.authority_records.append(DnsRRecord.decode(data, idx))
            except Exception as ex:
                logger.warning('Fail to decode authority record %d with error %s', i, str(ex))
        for i in range(additional_conunt):
            try:
                instance.additional_records.append(DnsRRecord.decode(data, idx))
            except Exception as ex:
                logger.warning('Fail to decode additional record %d with error %s', i, str(ex))
        return instance
    def encode(self, compress_names=True):
        '''
        :param compress_names: A flag that indicates if the domain names in the packed has to be compressed or not. Default to True
        :type compress_names: bool
        :return: A `bytearray` that represent the :class:`DnsPacket` instance.

        Encode the intance to a `bytearray` compressing domain names according with the given flag.
        '''
        raw_data = bytearray()
        if compress_names:
            name_compression_dictionary = {}
        else:
            name_compression_dictionary = None
        write_short(self._id, raw_data)
        write_short(self.__flags, raw_data)
        write_short(len(self.question_records), raw_data)
        write_short(len(self.answer_records), raw_data)
        write_short(len(self.authority_records), raw_data)
        write_short(len(self.additional_records), raw_data)
        for question_record in self.question_records:
            try:
                question_record.encode(raw_data, name_compression_dictionary)
            except Exception as ex:
                logger.warning('Fail to encode question record %s with error %s', question_record, str(ex))
        for answer_record in self.answer_records:
            try:
                answer_record.encode(raw_data, name_compression_dictionary)
            except Exception as ex:
                logger.warning('Fail to encode answer record %s with error %s', answer_record, str(ex))
        for authority_record in self.authority_records:
            try:
                authority_record.encode(raw_data, name_compression_dictionary)
            except Exception as ex:
                logger.warning('Fail to encode authority record %s with error %s', authority_record, str(ex))
        for additional_record in self.additional_records:
            try:
                additional_record.encode(raw_data, name_compression_dictionary)
            except Exception as ex:
                logger.warning('Fail to encode additional record %s with error %s', additional_record, str(ex))
        return bytes(raw_data)
    #utility methods
    def __get_flags_bit(self, bitmask):
        return self.__flags & bitmask != 0
    def __set_flags_bit(self, bitmask, value):
        if value:
            self.__flags |= bitmask
        else:
            self.__flags &= ~bitmask
    #properties
    @property
    def id(self):
        '''
        The query identifier of the packet.
        '''
        return self._id
    @id.setter
    def id(self, value):
        self._id = value
    @property
    def QR(self):
        '''
        Flag that indicates if the packet represent a query (True) or not (False).
        '''
        return not self.__get_flags_bit(self.__QR_MASK)
    @QR.setter
    def QR(self, value):
        self.__set_flags_bit(self.__QR_MASK, not value)
    @property
    def OpCode(self):
        '''
        The packet :class:`OpCode`.
        '''
        return ((self.__flags & self.__OPCODE_MASK) >> 11) & 0x0F
    @OpCode.setter
    def OpCode(self, value):
        self.__flags &= ~self.__OPCODE_MASK
        op_code = ((value.value) & 0x0F) << 11
        self.__flags |= op_code
    @property
    def AA(self):
        '''
        Flags that indicates if the packet is an authoritative answer (True) or not (False).
        '''
        return self.__get_flags_bit(self.__AA_MASK )
    @AA.setter
    def AA(self, value):
        self.__set_flags_bit(self.__AA_MASK, value)
    @property
    def TC(self):
        '''
        Flags that indicates if the packet is truncated (True) or not (False).
        '''
        return self.__get_flags_bit(self.__TC_MASK)
    @TC.setter
    def TC(self, value):
        self.__set_flags_bit(self.__TC_MASK, value)
    @property
    def RD(self):
        '''
        Flags that indicates if the name resolution recursion for the query is desired (True) or not (False).
        '''
        return self.__get_flags_bit(self.__RD_MASK)
    @RD.setter
    def RD(self, value):
        self.__set_flags_bit(self.__RD_MASK, value)
    @property
    def RA(self):
        '''
        Flags that indicates if the name resolution recursion is available (True) or not (False) on the server that send the reply.
        '''
        return self.__get_flags_bit(self.__RA_MASK)
    @RA.setter
    def RA(self, value):
        self.__set_flags_bit(self.__RA_MASK, value)
    @property
    def AD(self):
        '''
        Flags that indicates if the data in answer and authority sections has been authenticated (True) or not (False).
        '''
        return self.__get_flags_bit(self.__AD_MASK)
    @AD.setter
    def AD(self, value):
        self.__set_flags_bit(self.__AD_MASK, value)
    @property
    def CD(self):
        '''
        Flags that indicates if server that reply to the quesry has to authenticate the data in answer and authority sections (True) or not (False).
        '''
        return self.__get_flags_bit(self.__CD_MASK)
    @CD.setter
    def CD(self, value):
        self.__set_flags_bit(self.__CD_MASK, value)
    @property
    def RCODE(self):
        '''
        The packet :class:`RCode` (response code).
        '''
        return self.__flags & self.__RCODE_MASK
    @RCODE.setter
    def RCODE(self, value):
        self.__flags &= ~self.__RCODE_MASK
        self.__flags |= (value.value & 0x0F)
    @property
    def question_records(self):
        '''
        The list of question records in the packet.
        '''
        return self._question_records
    @property
    def answer_records(self):
        '''
        The list of answer records in the packet.
        '''
        return self._answer_records
    @property
    def authority_records(self):
        '''
        The list of authority records in the packet.
        '''
        return self._authority_records
    @property
    def additional_records(self):
        '''
        The list of additional records in the packet.
        '''
        return self._additional_records


from ._utility import Index, read_short, read_int, write_short, write_int
from ._dnsrecord import DnsRecord
from .dnsenums import DnsType, DnsClass
from socket import inet_ntoa, inet_aton

class DnsRRecord(DnsRecord):
    #subclass registry
    __subclass_registry = {}
    def __new__(cls, *args, **kwargs):
        if cls is DnsRRecord:
            raise Exception('Cannot instantiate abstract class {0}'.format(cls.__name__))
        else:
            return super().__new__(cls)
    def __init__(self, name, type_, class_, ttl):
        '''
        :param name: The record domain name.
        :type name: str
        :param type_: The record :class:`DnsType`.
        :type type_: int
        :param class_: The record :class:`DnsClass`.
        :type class_: int
        :param ttl: The record TTL in milliseconds.
        :type ttl: int

        Represent a DNS record and allows to maipulate its properties.
        This class is a base class for specialized record types and connot be instantiated.
        '''
        super().__init__(name, type_, class_)
        self._ttl = ttl
    #properties
    @property
    def ttl(self):
        '''
        The record TTL in milliseconds.
        '''
        return self._ttl
    @ttl.setter
    def ttl(self, value):
        self._ttl = value
    @property
    def flush_cache(self):
        '''
        A flag that indicates if the client that receive this record has to flush its cached copies (True) or not (False).
        '''
        return (self._class & 0x8000) != 0
    @flush_cache.setter
    def flush_cache(self, value):
        if value:
            self._class |= 0x8000
        else:
            self._class &= 0x7FFF
    #methods
    @staticmethod
    def register_subclass(record_type, subclass):
        if issubclass(subclass, DnsRRecord):
            DnsRRecord.__subclass_registry[record_type] = subclass
        else:
            raise RuntimeError('Decorated class must inherit from {0}'.format(DnsRRecord.__name__))
    @staticmethod
    def _get_record_type(data, idx):
        type_ = 0
        while True:
            step = data[idx]
            if step == 0:
                type_ = read_short(data, idx + 1)
                break
            elif step & 0xC0 != 0:
                idx.advance(2)
                type_ = read_short(data, idx)
                break
            else:
                idx.advance(step + 1)
        return type_
    @classmethod
    def decode(cls, data, idx):
        '''
        :param data: Array of bytes that has to be decoded.
        :type data: bytearray
        :param idx: The index from where the record data begins in the `data` array.
        :type idx: :class:`Index`

        Decode the given data starting from the given index into a proper :class:`DnsRRecord` subclass instance.
        '''
        if cls == DnsRRecord:
            type_ = DnsRRecord._get_record_type(data, Index(idx))
            if type_ in DnsRRecord.__subclass_registry:
                instance = DnsRRecord.__subclass_registry[type_].decode(data, idx)
            else:
                instance = DnsRRecordNotImplemented.decode(data, idx)
            return instance
        else:
            name, type_, class_ = super()._decode(data, idx)
            ttl = read_int(data, idx)
            data_size = read_short(data, idx)
            return name, type_, class_, ttl, data_size
    def encode(self, data, name_compression_dictionary):
        '''
        :param data: The `bytearray` to which the instance data will be appended.
        :type data: bytearray
        :param name_compression_dictionary: a dictionary used to perform name compression. Keys are the domain names, values are the indexes in the `data` array where the domain name begin.
        :type name_compression_dictionary: dict[str, int]

        Encode the intance to a `bytearray` compressing domain names according with the given `name_compression_dictionary`.
        The resulting `bytearray` will be attached to the given `data` parameter.
        If `name_compression_dictionary` is `None` no name compression will be applied.
        '''
        super()._encode(data, name_compression_dictionary)
        write_int(self.ttl, data)
    def answer_to(self, question):
        '''
        :param question: The :class:`DnsQRecord` instance for which the current instance has to be tested.
        :type question: :class:`DnsQRecord`
        :return: `True` if the instance answer to the given question, `False` otherwse.
        :rtype: bool

        Check if the record is a reply for the given question. A record answer to a question if all following criterias are satisfied:

        * The record and question domain names are the same.
        * The record and question types are the same OR the question type is :attr:`DnsType.ANY` OR the record type is :attr:`DnsType.CNAME`.
        * The record and question classes are the same OR the question class is :attr:`DnsClass.ANY`.
        '''
        #a record answer to a question only if they have:
        # - the same name
        # - the same type or question type is ANY or resource record type is CNAME
        # - the same class or the question class is ANY
        if not isinstance(question, DnsRecord):
            return False
        else:
            return (self.name == question.name and
                    (self.record_type == question.record_type or
                        question.record_type == DnsType.ANY or
                        self.record_type == DnsType.CNAME) and
                    (self.record_class == question.record_class or question.record_class == DnsClass.ANY))

#decorator for register DnsRRecord subclasses
def register_subclass(record_type):
    '''
    :param record_type: The :class:`DnsType` that the decorated class implements.
    :type record_type: :class:`DnsType`

    This decorator can only be applied to class that derives from :class:`DnsRRecord`.
    If applied to a class that does not satisfy the constrain a :exc:`RuntimeError` exception will be raised.
    Scope of this decorator is to register the class ad secoder for the given `record_type`.
    '''
    def worker(subclass):
        DnsRRecord.register_subclass(record_type, subclass)
        return subclass
    return worker

@register_subclass(DnsType.A)
class DnsRRecordA(DnsRRecord):
    def __init__(self, name, ttl, address):
        '''
        :param name: The record domain name.
        :type name: str
        :param ttl: The record TTL in milliseconds.
        :type ttl: int
        :param address: The record IP address (IPv4).
        :type address: str

        Represent a DNS Record of type A and allows to manipulate its properties.
        '''
        super().__init__(name, DnsType.A.value, DnsClass.IN.value, ttl)
        self.address = address
    #equality operator
    def __eq__(self, other):
        return (isinstance(other, DnsRRecordA) and
                super().__eq__(other) and
                self.address == other.address)
    #methods
    @classmethod
    def decode(cls, data, idx):
        name, _, class_, ttl, data_size = super().decode(data, idx)
        instance = DnsRRecordA(name, ttl, inet_ntoa(data[idx: idx + data_size]))
        #preserve cache flush bit
        instance._class = class_
        idx.advance(data_size)
        return instance
    def encode(self, data, name_compression_dictionary):
        super().encode(data, name_compression_dictionary)
        raw_ip = inet_aton(self.address)
        write_short(len(raw_ip), data)
        data.extend(raw_ip)

class DnsRRecordNotImplemented(DnsRRecord):
    def __init__(self, name, type_, class_, ttl, data_size, data):
        '''
        :param name: The record domain name.
        :type name: str
        :param type_: The record :class:`DnsType`.
        :type type_: int
        :param class_: The record :class:`DnsClass`.
        :type class_: int
        :param ttl: The record TTL in milliseconds.
        :type ttl: int
        :param data_size: The record data size in bytes.
        :type data_size: int
        :param data: The record raw data.
        :type data: bytearray

        Represent a DNS record and allows to maipulate its properties.
        This class is a base class for specialized record types and connot be instantiated.
        '''
        super().__init__(name, type_, class_, ttl)
        self._data_size = data_size
        self._data = data
    #equality operator
    def __eq__(self, other):
        return (isinstance(other, DnsRRecordNotImplemented) and
                super().__eq__(other) and
                self.data_size == other.data_size and
                self.data == other.data)
    #properties
    @property
    def data_size(self):
        '''
        The size of the record data section in bytes.
        '''
        return self._data_size
    @data_size.setter
    def data_size(self, value):
        self._data_size = value
    @property
    def data(self):
        '''
        The record data section (raw bytes representation).
        '''
        return self._data
    @data.setter
    def data(self, value):
        self._data = value        
    @property
    def is_unknown(self):
        '''
        A flag that indicates if the record represented by the instance is of a known :class:`DnsType` (True) or not (False).
        '''
        return not (DnsType.contains_value(self.record_type) and DnsClass.contains_value(self.record_class))
    #methods
    @classmethod
    def decode(cls, data, idx):
        name, type_, class_, ttl, data_size = super().decode(data, idx)
        instance = DnsRRecordNotImplemented(name, type_, class_, ttl, data_size, data[idx: idx + data_size])
        idx.advance(data_size)
        return instance
    def encode(self, data, name_compression_dictionary):
        super().encode(data, name_compression_dictionary)
        write_short(self.data_size, data)
        data.extend(self.data)


import setuptools

#get long description
with open('README.md') as readme:
    long_description = readme.read()
#get version number
version = '0.0.0a0'
with open('anubi/mdns/__init__.py') as info:
    for line in info:
        if '__version__' in line:
            version = line.split('=')[-1].strip(' \'\n')
            break


setuptools.setup(
    name='anubi.mdns',
    version=version,
    author='Andrea Parisotto',
    description='A simple pure python implementation of a multicast DNS responder',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Anubi85/mDNS',
    license='MIT',
    include_package_data=True,
    zip_safe=False,
    test_suite='tests',
    packages = setuptools.find_namespace_packages(include=['anubi.*']),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    keywords=['Multicast DNS', 'mDNS'],
    python_requires='>=3.8',
)


import unittest
from anubi.mdns import DnsRRecord, register_subclass, DnsType

class DecoratorTests(unittest.TestCase):

    def tearDown(self):
        if DnsType.TXT in DnsRRecord._DnsRRecord__subclass_registry:
            del DnsRRecord._DnsRRecord__subclass_registry[DnsType.TXT]
    def test_register_subclass_decorator_success(self):
        try:
            @register_subclass(DnsType.TXT)
            class TestClass(DnsRRecord):
                pass
        except RuntimeError as ex:
            if str(ex) == 'Decorated class must inherit from {0}'.format(DnsRRecord.__name__):
                self.fail() 
        self.assertTrue(DnsType.TXT in DnsRRecord._DnsRRecord__subclass_registry)
        self.assertEqual(DnsRRecord._DnsRRecord__subclass_registry[DnsType.TXT], TestClass)
    def test_register_subclass_decorator_failure(self):
        with self.assertRaises(RuntimeError) as context:
            @register_subclass(DnsType.TXT)
            class TestClass:
                pass
        self.assertEqual(str(context.exception), 'Decorated class must inherit from {0}'.format(DnsRRecord.__name__))
        self.assertFalse(DnsType.TXT in DnsRRecord._DnsRRecord__subclass_registry)
        with self.assertRaises(KeyError):
            _ = DnsRRecord._DnsRRecord__subclass_registry[DnsType.TXT]